@@ -1,5 +1,5 @@
 # -*- coding: iso-8859-1 -*-
-# Copyright (C) 2000-2005  Bastian Kleineidam
+# Copyright (C) 2000-2006 Bastian Kleineidam
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -18,7 +18,6 @@
 Handle for mailto: links.
 """
 
-import sys
 import cgi
 import urllib
 import smtplib
@@ -30,12 +29,37 @@ import linkcheck.log
 import linkcheck.dns.resolver
 
 
+def _split_address (address):
+    """
+    Split username and hostname of address. The hostname defaults
+    to 'localhost' if it is not specified.
+
+    @param address: an email address
+    @type address: string
+    @return: a tuple (username, hostname)
+    @rtype: tuple
+    @raise: LinkCheckerError if address could not be split
+    """
+    split = address.split("@", 1)
+    if len(split) == 2:
+        if not split[1]:
+            return (split[0], "localhost")
+        return tuple(split)
+    if len(split) == 1:
+        return (split[0], "localhost")
+    raise linkcheck.LinkCheckerError(_("Could not split the mail address"))
+
+
 class MailtoUrl (urlbase.UrlBase):
     """
     Url link with mailto scheme.
     """
 
     def build_url (self):
+        """
+        Call super.build_url(), extract list of mail addresses from URL,
+        and check their syntax.
+        """
         super(MailtoUrl, self).build_url()
         self.headers = {}
         self.addresses = email.Utils.getaddresses([self.cutout_addresses()])
@@ -46,13 +70,20 @@ class MailtoUrl (urlbase.UrlBase):
                     self.addresses.extend(email.Utils.getaddresses([a]))
         # check syntax of emails
         for name, addr in self.addresses:
-            username, domain = self._split_address(addr)
+            username, domain = _split_address(addr)
             if not linkcheck.url.is_safe_domain(domain):
                 raise linkcheck.LinkCheckerError(_("Invalid mail syntax"))
-        linkcheck.log.debug(linkcheck.LOG_CHECK, "addresses: %s",
-                            self.addresses)
+        assert None == linkcheck.log.debug(linkcheck.LOG_CHECK,
+            "addresses: %s", self.addresses)
 
     def cutout_addresses (self):
+        """
+        Parse all mail addresses out of the URL target. Additionally
+        store headers.
+
+        @return: comma separated list of email addresses
+        @rtype: string
+        """
         # cut off leading mailto: and unquote
         url = urllib.unquote(self.base_url[7:])
         # search for cc, bcc, to and store in headers
@@ -76,10 +107,12 @@ class MailtoUrl (urlbase.UrlBase):
             elif mode == 2:
                 mode = 0
         if i < (len(url) - 1):
-            headers = cgi.parse_qs(url[(i+1):], strict_parsing=True)
-            for key, val in headers.items():
-                key = key.lower()
-                self.headers.setdefault(key, []).extend(val)
+            try:
+                headers = cgi.parse_qs(url[(i+1):], strict_parsing=True)
+                for key, val in headers.iteritems():
+                    self.headers.setdefault(key.lower(), []).extend(val)
+            except ValueError, err:
+                self.add_warning(_("Error parsing CGI values: %s") % str(err))
             addrs = url[:i]
         else:
             addrs = url
@@ -102,7 +135,8 @@ class MailtoUrl (urlbase.UrlBase):
              If not, print a warning.
         """
         if not self.addresses:
-            self.add_warning(_("No addresses found."))
+            self.add_warning(_("No addresses found."),
+                             tag="mail-no-addresses")
             return
         for name, mail in self.addresses:
             self.check_smtp_domain(name, mail)
@@ -112,65 +146,83 @@ class MailtoUrl (urlbase.UrlBase):
         """
         Check a single mail address.
         """
-        linkcheck.log.debug(linkcheck.LOG_CHECK,
-                            "checking mail address %r", mail)
-        linkcheck.log.debug(linkcheck.LOG_CHECK, "splitting address")
-        username, domain = self._split_address(mail)
-        linkcheck.log.debug(linkcheck.LOG_CHECK,
-                            "looking up MX mailhost %r", domain)
+        assert None == linkcheck.log.debug(linkcheck.LOG_CHECK,
+            "checking mail address %r", mail)
+        assert None == linkcheck.log.debug(linkcheck.LOG_CHECK,
+            "splitting address")
+        username, domain = _split_address(mail)
+        assert None == linkcheck.log.debug(linkcheck.LOG_CHECK,
+            "looking up MX mailhost %r", domain)
         answers = linkcheck.dns.resolver.query(domain, 'MX')
         if len(answers) == 0:
-            self.add_warning(_("No MX mail host for %(domain)s found.") % \
-                            {'domain': domain})
-            return
-        # sort according to preference (lower preference means this
-        # host should be preferred
-        mxdata = [(rdata.preference,
-                   rdata.exchange.to_text(omit_final_dot=True))
-                   for rdata in answers]
-        mxdata.sort()
+            self.add_warning(_("No MX mail host for %(domain)s found.") %
+                            {'domain': domain},
+                             tag="mail-no-mx-host")
+            answers = linkcheck.dns.resolver.query(domain, 'A')
+            if len(answers) == 0:
+                self.set_result(_("No host for %(domain)s found.") %
+                                 {'domain': domain}, valid=False)
+                return
+            # set preference to zero
+            mxdata = [(0, rdata.to_text(omit_final_dot=True))
+                      for rdata in answers]
+        else:
+            mxdata = [(rdata.preference,
+                       rdata.exchange.to_text(omit_final_dot=True))
+                       for rdata in answers]
+            # sort according to preference (lower preference means this
+            # host should be preferred)
+            mxdata.sort()
         # debug output
-        linkcheck.log.debug(linkcheck.LOG_CHECK,
-                            "found %d MX mailhosts:", len(answers))
+        assert None == linkcheck.log.debug(linkcheck.LOG_CHECK,
+            "found %d MX mailhosts:", len(answers))
         for preference, host in mxdata:
-            linkcheck.log.debug(linkcheck.LOG_CHECK,
-                                "MX host %r, preference %d", host, preference)
+            assert None == linkcheck.log.debug(linkcheck.LOG_CHECK,
+                "MX host %r, preference %d", host, preference)
         # connect
-        self.check_smtp_connect(mxdata, username)
+        self.check_smtp_connect(mxdata, username, domain)
 
-    def check_smtp_connect (self, mxdata, username):
+    def check_smtp_connect (self, mxdata, username, domain):
         """
-        mxdata is a list of (preference, host) tuples to check for
+        Connect to SMTP servers and check emails.
+
+        @param mxdata: list of (preference, host) tuples to check for
+        @type mxdata: list
+        @param username: the username to verify
+        @type username: string
         """
         smtpconnect = 0
         for preference, host in mxdata:
             try:
-                linkcheck.log.debug(linkcheck.LOG_CHECK,
-                        "SMTP check for %r (preference %d)", host, preference)
+                assert None == linkcheck.log.debug(linkcheck.LOG_CHECK,
+                    "SMTP check for %r (preference %d)", host, preference)
                 self.url_connection = smtplib.SMTP()
-                if self.consumer.config.get("debug"):
+                if linkcheck.log.is_debug(linkcheck.LOG_CHECK):
                     self.url_connection.set_debuglevel(1)
                 self.url_connection.connect(host)
-                linkcheck.log.debug(linkcheck.LOG_CHECK, "SMTP connected!")
+                assert None == linkcheck.log.debug(linkcheck.LOG_CHECK,
+                    "SMTP connected!")
                 smtpconnect = 1
                 self.url_connection.helo()
-                info = self.url_connection.verify(username)
-                linkcheck.log.debug(linkcheck.LOG_CHECK,
-                                    "SMTP user info %r", info)
+                info = self.url_connection.verify("%s@%s" % (username, domain))
+                assert None == linkcheck.log.debug(linkcheck.LOG_CHECK,
+                    "SMTP user info %r", info)
                 d = {'info': str(info[1])}
                 if info[0] == 250:
                     self.add_info(_("Verified address: %(info)s.") % d)
                 # check for 25x return code which means that the address
                 # could not be verified, but is sent anyway
                 elif 0 < (info[0] - 250) < 10:
-                    self.add_info(_("Unverified address: %(info)s." \
+                    self.add_info(_("Unverified address: %(info)s."
                                   " But mail will be sent anyway.") % d)
                 else:
-                    self.add_warning(_("Unverified address: %(info)s.") % d)
+                    self.add_warning(_("Unverified address: %(info)s.") % d,
+                     tag="mail-unverified-address")
             except smtplib.SMTPException, msg:
                 self.add_warning(
-                      _("MX mail host %(host)s did not accept connections: " \
-                        "%(error)s.") % {'host': host, 'error': str(msg)})
+                      _("MX mail host %(host)s did not accept connections: "
+                        "%(error)s.") % {'host': host, 'error': str(msg)},
+                        tag="mail-no-connection")
             if smtpconnect:
                 break
         if not smtpconnect:
@@ -178,17 +230,6 @@ class MailtoUrl (urlbase.UrlBase):
         else:
             self.set_result(_("Found MX mail host %(host)s") % {'host': host})
 
-    def _split_address (self, address):
-        split = address.split("@", 1)
-        if len(split) == 2:
-            if not split[1]:
-                return (split[0], "localhost")
-            return tuple(split)
-        if len(split) == 1:
-            return (split[0], "localhost")
-        raise linkcheck.LinkCheckerError(
-                                  _("Could not split the mail address"))
-
     def close_connection (self):
         """
         Close a possibly opened SMTP connection.
@@ -198,7 +239,8 @@ class MailtoUrl (urlbase.UrlBase):
             return
         try:
             self.url_connection.quit()
-        except smtplib.SMTPException:
+        except:
+            # ignore close errors
             pass
         self.url_connection = None
 
@@ -217,5 +259,6 @@ class MailtoUrl (urlbase.UrlBase):
         mailto: URLs do not have any content
 
         @return: False
+        @rtype: bool
         """
         return False
