@@ -22,10 +22,11 @@
 import socket
 import sets
 import sys
-import os
 import time
 import encodings.idna
 
+import linkcheck
+import linkcheck.log
 import linkcheck.dns.exception
 import linkcheck.dns.message
 import linkcheck.dns.name
@@ -55,14 +56,28 @@ class NoNameservers(linkcheck.dns.exception.DNSException):
     """No non-broken nameservers are available to answer the query."""
     pass
 
+class NotAbsolute(linkcheck.dns.exception.DNSException):
+    """Raised if an absolute domain name is required but a relative name
+    was provided."""
+    pass
+
+class NoRootSOA(linkcheck.dns.exception.DNSException):
+    """Raised if for some reason there is no SOA at the root name.
+    This should never happen!"""
+    pass
+
+
 class Answer(object):
     """DNS stub resolver answer
 
     Instances of this class bundle up the result of a successful DNS
     resolution.
 
-    For convenience, the answer is iterable.  "for a in answer" is
-    equivalent to "for a in answer.rrset".
+    For convenience, the answer object implements much of the sequence
+    protocol, forwarding to its rrset.  E.g. "for a in answer" is
+    equivalent to "for a in answer.rrset", "answer[i]" is equivalent
+    to "answer.rrset[i]", and "answer[i:j]" is equivalent to
+    "answer.rrset[i:j]".
 
     Note that CNAMEs or DNAMEs in the response may mean that answer
     node's name might not be the query name.
@@ -108,13 +123,16 @@ class Answer(object):
                             break
                         continue
                     except KeyError:
-                        raise NoAnswer("DNS response had no answer")
-                raise NoAnswer("DNS response had no answer")
+                        raise NoAnswer, "DNS response had no answer"
+                raise NoAnswer, "DNS response had no answer"
         if rrset is None:
-            raise NoAnswer("DNS response had no answer")
+            raise NoAnswer, "DNS response had no answer"
         self.rrset = rrset
         self.expiration = time.time() + min_ttl
 
+    def __str__ (self):
+        return str(self.rrset)
+
     def __getattr__(self, attr):
         if attr == 'name':
             return self.rrset.name
@@ -135,6 +153,19 @@ class Answer(object):
     def __iter__(self):
         return iter(self.rrset)
 
+    def __getitem__(self, i):
+        return self.rrset[i]
+
+    def __delitem__(self, i):
+        del self.rrset[i]
+
+    def __getslice__(self, i, j):
+        return self.rrset[i:j]
+
+    def __delslice__(self, i, j):
+        del self.rrset[i:j]
+
+
 class Cache(object):
     """Simple DNS answer cache.
 
@@ -285,6 +316,7 @@ class Resolver(object):
           'localhost',
           'loopback',
           '127.0.0.1',
+          '0.0.0.0',
           '::1',
           'ip6-localhost',
           'ip6-loopback',
@@ -334,20 +366,24 @@ class Resolver(object):
             self.nameservers.append('127.0.0.1')
 
     def read_local_hosts (self):
-        # XXX is this default list of localhost stuff complete?
         self.add_addrinfo(socket.gethostname())
         # add system specific hosts for all enabled interfaces
+        self.add_addrinfo('localhost', interface=True)
         for addr in self.read_local_ifaddrs():
             self.add_addrinfo(addr, interface=True)
 
     def read_local_ifaddrs (self):
-        """all active interfaces' ip addresses"""
+        """
+        IP addresses for all active interfaces.
+        @return: list of IP addresses
+        @rtype: list of strings
+        """
         if not sys.platform.startswith('linux'):
             # only Linux is supported right now
             return []
         import linkcheck.dns.ifconfig
         ifc = linkcheck.dns.ifconfig.IfConfig()
-        return [ifc.getAddr(iface) for iface in ifc.getInterfaceList()
+        return [ifc.getAddr(iface) for iface in ifc.getInterfaceList() \
                 if ifc.isUp(iface)]
 
     def add_addrinfo (self, host, interface=False):
@@ -492,6 +528,22 @@ class Resolver(object):
         finally:
             lm.Close()
 
+    def _compute_timeout(self, start):
+        now = time.time()
+        if now < start:
+            if start - now > 1:
+                # Time going backwards is bad.  Just give up.
+                raise Timeout
+            else:
+                # Time went backwards, but only a little. This can
+                # happen, e.g. under vmware with older linux kernels.
+                # Pretend it didn't happen.
+                now = start
+        duration = now - start
+        if duration >= self.lifetime:
+            raise Timeout
+        return min(self.lifetime - duration, self.timeout)
+
     def query(self, qname, rdtype=linkcheck.dns.rdatatype.A,
               rdclass=linkcheck.dns.rdataclass.IN, tcp=False):
         """Query nameservers to find the answer to the question.
@@ -549,22 +601,18 @@ class Resolver(object):
                 request.use_tsig(self.keyring, self.keyname)
             request.use_edns(self.edns, self.ednsflags, self.payload)
             response = None
-            #
+
             # make a copy of the servers list so we can alter it later.
-            #
+
             nameservers = self.nameservers[:]
+            backoff = 0.10
             while response is None:
                 if len(nameservers) == 0:
-                    raise NoNameservers("No DNS servers could answer the query")
-                for nameserver in nameservers:
-                    now = time.time()
-                    if now < start:
-                        # Time going backwards is bad.  Just give up.
-                        raise Timeout("DNS query timed out")
-                    duration = now - start
-                    if duration >= self.lifetime:
-                        raise Timeout("DNS query timed out")
-                    timeout = min(self.lifetime - duration, self.timeout)
+                    raise NoNameservers, \
+                      "No DNS servers %s could answer the query %s" % \
+                      (str(self.nameservers), str(qname))
+                for nameserver in nameservers[:]:
+                    timeout = self._compute_timeout(start)
                     try:
                         if tcp:
                             response = linkcheck.dns.query.tcp(request, nameserver,
@@ -573,24 +621,24 @@ class Resolver(object):
                             response = linkcheck.dns.query.udp(request, nameserver,
                                                      timeout, self.port)
                     except (socket.error, linkcheck.dns.exception.Timeout):
-                        #
+
                         # Communication failure or timeout.  Go to the
                         # next server
-                        #
+
                         response = None
                         continue
                     except linkcheck.dns.query.UnexpectedSource:
-                        #
+
                         # Who knows?  Keep going.
-                        #
+
                         response = None
                         continue
                     except linkcheck.dns.exception.FormError:
-                        #
+
                         # We don't understand what this server is
                         # saying.  Take it out of the mix and
                         # continue.
-                        #
+
                         nameservers.remove(nameserver)
                         response = None
                         continue
@@ -598,13 +646,34 @@ class Resolver(object):
                     if rcode == linkcheck.dns.rcode.NOERROR or \
                            rcode == linkcheck.dns.rcode.NXDOMAIN:
                         break
+
+                    # We got a response, but we're not happy with the
+                    # rcode in it.  Remove the server from the mix if
+                    # the rcode isn't SERVFAIL.
+
+                    if rcode != linkcheck.dns.rcode.SERVFAIL:
+                        nameservers.remove(nameserver)
                     response = None
+
+                if response is not None:
+                    break
+                # All nameservers failed!
+
+                if len(nameservers) > 0:
+
+                    # But we still have servers to try.  Sleep a bit
+                    # so we don't pound them!
+
+                    timeout = self._compute_timeout(start)
+                    sleep_time = min(timeout, backoff)
+                    backoff *= 2
+                    time.sleep(sleep_time)
             if response.rcode() == linkcheck.dns.rcode.NXDOMAIN:
                 continue
             all_nxdomain = False
             break
         if all_nxdomain:
-            raise NXDOMAIN("Domain does not exist")
+            raise NXDOMAIN, "Domain does not exist"
         answer = Answer(qname, rdtype, rdclass, response)
         if self.cache:
             self.cache.put((qname, rdtype, rdclass), answer)
@@ -645,15 +714,52 @@ class Resolver(object):
 
 default_resolver = None
 
+def get_default_resolver ():
+    global default_resolver
+    if default_resolver is None:
+        default_resolver = Resolver()
+    return default_resolver
+
 def query(qname, rdtype=linkcheck.dns.rdatatype.A, rdclass=linkcheck.dns.rdataclass.IN,
-          tcp=False):
+          tcp=False, resolver=None):
     """Query nameservers to find the answer to the question.
 
     This is a convenience function that uses the default resolver
     object to make the query.
     @see: L{linkcheck.dns.resolver.Resolver.query} for more information on the
     parameters."""
-    global default_resolver
-    if default_resolver is None:
-        default_resolver = Resolver()
-    return default_resolver.query(qname, rdtype, rdclass, tcp)
+    assert None == linkcheck.log.debug(linkcheck.LOG_DNS,
+        "Query %s %s %s", qname, rdtype, rdclass)
+    if resolver is None:
+        resolver = get_default_resolver()
+    return resolver.query(qname, rdtype, rdclass, tcp)
+
+def zone_for_name(name, rdclass=linkcheck.dns.rdataclass.IN,
+                  tcp=False, resolver=None):
+    """Find the name of the zone which contains the specified name.
+
+    @param name: the query name
+    @type name: absolute linkcheck.dns.name.Name object or string
+    @ivar rdclass: The query class
+    @type rdclass: int
+    @param tcp: use TCP to make the query (default is False).
+    @type tcp: bool
+    @param resolver: the resolver to use
+    @type resolver: linkcheck.dns.resolver.Resolver object or None
+    @rtype: dns.name.Name"""
+
+    if isinstance(name, str):
+        name = linkcheck.dns.name.from_text(name, linkcheck.dns.name.root)
+    if resolver is None:
+        resolver = get_default_resolver()
+    if not name.is_absolute():
+        raise NotAbsolute, name
+    while 1:
+        try:
+            answer = resolver.query(name, linkcheck.dns.rdatatype.SOA, rdclass, tcp)
+            return name
+        except (linkcheck.dns.resolver.NXDOMAIN, linkcheck.dns.resolver.NoAnswer):
+            try:
+                name = name.parent()
+            except linkcheck.dns.name.NoParent:
+                raise NoRootSOA
